@@ -19,7 +19,7 @@
 ;; URL:              https://github.com/meedstrom/el-job
 ;; Created:          2024-10-30
 ;; Keywords:         processes
-;; Package-Version:  2.4.7
+;; Package-Version:  2.4.8
 ;; Package-Requires: ((emacs "29.1"))
 
 ;;; Commentary:
@@ -32,7 +32,7 @@
 ;; result as if it had been returned by that `mapcar'.  In the meantime,
 ;; current Emacs does not hang at all.
 
-;; You need to know the concept of a callback.
+;; You do need to grok the concept of a callback.
 
 ;; Public API:
 ;; - Function `el-job-launch' (main entry point)
@@ -61,9 +61,6 @@
 (defvar el-job-major-version 2
   "Number incremented for breaking changes.")
 
-(define-obsolete-variable-alias 'el-jobs 'el-job--all-jobs "2.2.0 (2025-03-09)"
-  "I liked this name, but it breaks an Elisp convention.")
-
 
 ;;; Subroutines
 
@@ -120,8 +117,8 @@ an .eln anyway, without your having to recompile on save."
                               ;; `native-comp-unit-file'.  Current workaround
                               ;; is return nil, ie fall back on FILE; not
                               ;; ideal.
-                              (when (file-exists-p eln)
-                                eln))))
+                              (and (file-exists-p eln)
+                                   eln))))
        file)))
 
 (defvar el-job--onetime-canary nil)
@@ -155,16 +152,16 @@ find the correct file."
             (locate-file (symbol-name feature) load-path '(".el" ".el.gz"))))
     (if (or (string-suffix-p ".el" loaded)
             (string-suffix-p ".el.gz" loaded))
-        (or (when (native-comp-available-p)
-              ;; If we built an .eln last time, return it now even though the
-              ;; current Emacs process is still running interpreted .el.
-              ;;
-              ;; NOTE: Thanks to hashing file contents, `comp-lookup-eln'
-              ;; returns nil if the .el has changed on disk, even if the
-              ;; developer did not eval-buffer again.  Then we proceed to build
-              ;; another .eln.  Not sure I'm a fan... but eh, Works For Me, and
-              ;; it removes the need to `eval-buffer' constantly.
-              (comp-lookup-eln loaded))
+        (or (and (native-comp-available-p)
+                 ;; If we built an .eln last time, return it now even though
+                 ;; the current Emacs process is still running interpreted .el.
+                 ;;
+                 ;; NOTE: Thanks to hashing file contents, `comp-lookup-eln'
+                 ;; returns nil if the .el has changed on disk, even if the
+                 ;; developer did not eval-buffer again.  Then we proceed to
+                 ;; build another .eln.  Not sure I'm a fan... but eh, Works
+                 ;; For Me, and removes the need to `eval-buffer' constantly.
+                 (comp-lookup-eln loaded))
             (let* ((elc (file-name-concat temporary-file-directory
                                           (concat (symbol-name feature)
                                                   ".elc")))
@@ -277,9 +274,8 @@ being saddled with a huge item in addition to the average workload."
                     (push item items)))))))
         (if (length= sublists 0)
             (progn
-              ;; Degrade gracefully (but I think we never arrive here)
               (fset 'el-job--split-optimally #'el-job--split-evenly)
-              (cl-assert (not (length= sublists 0))))
+              (error "Internal coding mistake, degrading gracefully from now"))
           ;; Spread leftovers equally
           (let ((ctr 0)
                 (len (length sublists)))
@@ -297,9 +293,25 @@ See subroutine `el-job-child--zip' for details."
       (setq merged (el-job-child--zip (pop lists) merged)))
     merged))
 
+(defun el-job--windows-cores ()
+  "Try to get number of physical cores on Windows system."
+  (with-temp-buffer
+    (call-process "cmd.exe" nil t nil "/C" "wmic CPU get NumberOfCores /value")
+    (goto-char 1)
+    (or (and (re-search-forward "NumberOfCores=\\([0-9]+\\)" nil t)
+             (string-to-number (match-string 1)))
+        1)))
+
 
 ;;; Main logic
 
+(defcustom el-job-max-cores nil
+  "A limit on the number of subprocesses for one job.
+Windows can get \"error: Could not create child process\" if making too
+many processes, so capping it can help."
+  :type '(choice integer (const :tag "Auto" nil))
+  :group 'processes)
+
 (defvar el-job--all-jobs (make-hash-table :test #'eq)
   "Table of all el-job objects.")
 
@@ -393,8 +405,8 @@ INPUTS is a list that will be split by up to the output of
 INPUTS can also be a function that returns a list.  In this case, the
 function is deferred until needed, possibly saving on compute.
 
-If INPUTS returns nil, do nothing and return the symbol
-`inputs-were-empty'.
+If INPUTS is a function and it returns nil, do nothing and return the
+symbol `inputs-were-empty'.
 
 
 The subprocesses have no access to current Emacs state.  The only way
@@ -439,6 +451,8 @@ For debugging, see these commands:
   (when callback
     (unless (and (symbolp callback)
                  (functionp callback))
+      ;; This one doesn't actually need to be a symbol, but easier to debug
+      ;; and user is not surprised since it seems consistent
       (error "Argument CALLBACK must be a symbol with a function definition")))
   (unless (proper-list-p load-features)
     (error "Argument LOAD-FEATURES must be a list"))
@@ -470,10 +484,14 @@ For debugging, see these commands:
           (when do-exec
             (setf .callback callback)
             ;; Prevent spawning a dozen processes when we need only one or two
-            (let ((machine-cores (max 1 (1- (num-processors)))))
-              (setf .n-cores-to-use (if (length< .queued-inputs machine-cores)
-                                        (length .queued-inputs)
-                                      machine-cores))
+            (let ((machine-cores (max 1 (or el-job-max-cores
+                                            (and (eq system-type 'windows-nt)
+                                                 (- (el-job--windows-cores) 1))
+                                            (- (num-processors) 1)))))
+              (setf .n-cores-to-use
+                    (if (length< .queued-inputs machine-cores)
+                        (length .queued-inputs)
+                      machine-cores))
               (when (or (length< .ready .n-cores-to-use)
                         (not (cl-every #'process-live-p .ready)))
                 (setq do-respawn t)))
@@ -529,33 +547,34 @@ see `el-job-launch'."
               (setq-local el-job-here job)
               (erase-buffer)
               (current-buffer)))
-      (condition-case err
-          (dotimes (i .n-cores-to-use)
-            (let ((proc (make-process
-                         :name (format "el-job:%s:%d" .id i)
-                         :noquery t
-                         :connection-type 'pipe
-                         ;; https://github.com/jwiegley/emacs-async/issues/165
-                         :coding 'utf-8-emacs-unix
-                         :stderr .stderr
-                         :buffer (get-buffer-create
-                                  (format " *el-job:%s:%d*" .id i) t)
-                         :command command
-                         :sentinel #'ignore)))
-              (when (string-suffix-p ">" (process-name proc))
-                (el-job--dbg 1 "Unintended duplicate process id for %s" proc))
-              (with-current-buffer (process-buffer proc)
-                (setq-local el-job-here job)
-                (process-send-string proc vars)
-                (process-send-string proc "\n")
-                (process-send-string proc libs)
-                (process-send-string proc "\n"))
-              (push proc .ready)))
-        ;; https://github.com/meedstrom/org-node/issues/75
-        (( file-error )
-         (el-job--disable job)
-         (el-job--dbg 1 "Terminated job because of: %S" err)
-         (setq return-value err)))
+      (with-temp-buffer ;; https://github.com/meedstrom/org-node/issues/98
+        (condition-case err
+            (dotimes (i .n-cores-to-use)
+              (let ((proc (make-process
+                           :name (format "el-job:%s:%d" .id i)
+                           :noquery t
+                           :connection-type 'pipe
+                           ;; https://github.com/jwiegley/emacs-async/issues/165
+                           :coding 'utf-8-emacs-unix
+                           :stderr .stderr
+                           :buffer (get-buffer-create
+                                    (format " *el-job:%s:%d*" .id i) t)
+                           :command command
+                           :sentinel #'ignore)))
+                (when (string-suffix-p ">" (process-name proc))
+                  (el-job--dbg 1 "Unintended duplicate process id for %s" proc))
+                (with-current-buffer (process-buffer proc)
+                  (setq-local el-job-here job)
+                  (process-send-string proc vars)
+                  (process-send-string proc "\n")
+                  (process-send-string proc libs)
+                  (process-send-string proc "\n"))
+                (push proc .ready)))
+          ;; https://github.com/meedstrom/org-node/issues/75
+          (( file-error )
+           (el-job--disable job)
+           (el-job--dbg 1 "Terminated job because of: %S" err)
+           (setq return-value err))))
       ;; Return non-nil on error, so caller can choose to fail quietly.
       ;; We suppressed the error signal because for some users, it only occurs
       ;; intermittently and does not break things.
@@ -570,6 +589,8 @@ should trigger `el-job--handle-output'."
   (el-job--with job
       ( .ready .busy .input-sets .result-sets .queued-inputs .n-cores-to-use
         .past-elapsed .timestamps .finish-times .id .stderr .timer )
+    (when .busy
+      (error "Unexpected still-busy processes at this time: %S" .busy))
     (cancel-timer .timer)
     (setf .input-sets nil)
     (setf .result-sets nil)
@@ -578,26 +599,32 @@ should trigger `el-job--handle-output'."
                                            .n-cores-to-use
                                            .past-elapsed))
           busy-bufs)
-      (unless (length< splits (1+ (length .ready)))
-        (error "el-job: Items split in %d lists, but only %d ready processes"
-               (length splits) (length .ready)))
+      ;; Sanity check
+      (when (length> splits (length .ready))
+        (if (or (null .ready) (null splits))
+            (error "el-job: Items split in %d lists, but only %d ready processes "
+                   (length splits) (length .ready))
+          (warn "el-job: Items split in %d lists, but only %d ready processes. %s"
+                (length splits) (length .ready)
+                "Falling back to use 1 process")
+          (mapc #'delete-process (cdr .ready))
+          (setf .ready (list (car .ready)))
+          (setq splits (list .queued-inputs))))
       (let ((print-length nil)
             (print-level nil)
             (print-circle t)
             (print-symbols-bare t)
             (print-escape-newlines t)
-            items proc)
-        (while (progn
-                 (setq items (pop splits))
-                 (setq proc (pop .ready))
-                 (push proc .busy)
-                 (push (process-buffer proc) busy-bufs)
-                 (setf (alist-get proc .input-sets) items)
-                 (with-current-buffer (process-buffer proc)
-                   (erase-buffer)
-                   (process-send-string proc (prin1-to-string items))
-                   (process-send-string proc "\n"))
-                 splits)))
+            proc)
+        (dolist (items splits)
+          (setq proc (pop .ready))
+          (push proc .busy)
+          (push (process-buffer proc) busy-bufs)
+          (setf (alist-get proc .input-sets) items)
+          (with-current-buffer (process-buffer proc)
+            (erase-buffer)
+            (process-send-string proc (prin1-to-string items))
+            (process-send-string proc "\n"))))
       (setf .queued-inputs nil)
       (plist-put .timestamps :work-begun (current-time))
       (setf .timer (run-with-timer .02 nil #'el-job--poll 1 busy-bufs)))))
@@ -659,13 +686,12 @@ after a short delay.  N is the count of checks done so far."
       (el-job--disable job)
       (el-job--dbg 2 "Reaped idle processes for %s" (el-job-id job)))))
 
-;; TODO: Consider setting `inhibit-quit' to nil or whatever is needed to let
-;;       us edebug the callback
+;; REVIEW: Consider setting `inhibit-quit' to nil (called by timer means t now)
 (defun el-job--handle-output ()
   "Handle output in current buffer.
 
 If this is the last output for the job, merge all outputs, maybe execute
-the callback function, finally maybe run the job again if there is now
+the callback function, finally maybe run the job again if there is
 more input in the queue."
   (let ((proc (get-buffer-process (current-buffer)))
         (job el-job-here)
@@ -704,8 +730,10 @@ more input in the queue."
         ;; Somehow, it took 700 lines of code to get here.
         (when .callback
           ;; Allow quitting out of a hung or slow CALLBACK.
+          ;; TODO: Recover well.  How?
           (with-local-quit
-            (funcall .callback (el-job--zip-all .result-sets) job)))
+            (funcall .callback (el-job--zip-all .result-sets) job)
+            t))
         ;; REVIEW: Either
         ;; 1. get rid of this code path (see comments in `el-job-launch' re.
         ;;    queued inputs); or
@@ -783,7 +811,7 @@ for buffer names starting with \" *el-job\" - note leading space."
              (remhash id el-job--all-jobs))
            el-job--all-jobs))
 
-(defmacro el-job--sit-until-nil-p (test max-secs &optional message)
+(defmacro el-job--sit-until-not (test max-secs &optional message)
   "Block until form TEST evaluates to nil, or time MAX-SECS has elapsed.
 Either way, return the last TEST result, i.e. non-nil if timed out.
 
@@ -813,34 +841,45 @@ A typical TEST would check if something in the environment has changed."
 If the job has still not finished after MAX-SECS seconds, stop
 blocking and return nil.
 
-Meanwhile, ensure string MESSAGE is visible in the minibuffer."
-  (not (el-job--sit-until-nil-p (el-job-is-busy id) max-secs message)))
+Meanwhile, ensure string MESSAGE is visible in the minibuffer.
+
+If there is no job that matches ID, immediately return t."
+  (when (el-job-p id)
+    (error "el-job-is-busy: Passed a job object, but expected only a job ID"))
+  (let ((job (gethash id el-job--all-jobs)))
+    (if job
+        (not (el-job--sit-until-not (el-job-busy job) max-secs message))
+      t)))
 
 (defun el-job-is-busy (id)
   "Return list of busy processes for job ID, if any.
 Safely return nil otherwise, whether or not ID is known."
   (when (el-job-p id)
-    (error "el-job-is-busy: Passed a job object, but expected only an ID"))
+    (error "el-job-is-busy: Passed a job object, but expected only a job ID"))
   (let ((job (gethash id el-job--all-jobs)))
     (and job (el-job-busy job))))
 
 
-;;; Ok, don't need to break convention anymore.
+;;; Obsolete
 
-(define-obsolete-function-alias 'el-job:id             #'el-job-id             "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:callback       #'el-job-callback       "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:n-cores-to-use #'el-job-n-cores-to-use "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:ready          #'el-job-ready          "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:busy           #'el-job-busy           "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:stderr         #'el-job-stderr         "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:timestamps     #'el-job-timestamps     "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:poll-timer     #'el-job-timer          "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:finish-times   #'el-job-finish-times   "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:spawn-args     #'el-job-spawn-args     "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:past-elapsed   #'el-job-past-elapsed   "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:queued-inputs  #'el-job-queued-inputs  "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:input-sets     #'el-job-input-sets     "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:result-sets    #'el-job-result-sets    "2.3.0 (2025-03-16)")
+;; Accelerate the timeframe until we can clean the namespace
+(defvar el-jobs :obsolete)
+(let ((complainer
+       (lambda (_) (error "Some renames in el-job 2.3.0, update your code"))))
+  (fset 'el-job:id             complainer)
+  (fset 'el-job:callback       complainer)
+  (fset 'el-job:n-cores-to-use complainer)
+  (fset 'el-job:ready          complainer)
+  (fset 'el-job:busy           complainer)
+  (fset 'el-job:stderr         complainer)
+  (fset 'el-job:timestamps     complainer)
+  (fset 'el-job:poll-timer     complainer)
+  (fset 'el-job:finish-times   complainer)
+  (fset 'el-job:spawn-args     complainer)
+  (fset 'el-job:past-elapsed   complainer)
+  (fset 'el-job:queued-inputs  complainer)
+  (fset 'el-job:input-sets     complainer)
+  (fset 'el-job:result-sets    complainer))
 
 (provide 'el-job)
 
